% PPTRANGE Calculate bounds on steering inequality for PPT states
%
%  Based on <a href="http://arxiv.org/abs/1305.1767">arXiv:1305.1767</a>, eq. 9
%
%  [maxval,minval] = PPTRANGE(ineqops,level) takes the parameters:
%
%   ineqops: size [$d_B$, $d_B$, $n_A$, $m_A$], collection of F matrices
%   level, aka $l$: Level of hierarchy
%
%  And attempts to solve Equation (9) plus the corresponding minimization
%
%  Requires chi_validity.m, Tx.m, and a working YALMIP setup
%
%  Latest version at
%  <a href="https://www.mattpusey.uk/steer/">www.mattpusey.co.uk/steer</a>,
%  bug reports welcome at <a href="mailto:m@physics.org">m@physics.org</a>

% Copyright (C) 2013-2014 Matthew F. Pusey
%
% This program is free software; you can redistribute it and/or
% modify it under the terms of the GNU General Public License
% as published by the Free Software Foundation; either version 2
% of the License, or (at your option) any later version.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
%
% You should have received a copy of the GNU General Public License
% along with this program; if not, write to the Free Software
% Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
% MA 02110-1301, USA.

function [maxval,minval] = PPTrange(ineqops,level)
[~,d,outcomes,settings] = size(ineqops);
[outsize,xpart,num_can,can_outcome,can_setting,can_herm] ...
    = chi_validity(outcomes, settings, level);

x = zeros(outsize*d);
u = cell(num_can);
val = 0;

for i=1:num_can
    if can_herm{i}
        type = 'hermitian';
    else
        type = 'full';
    end
    u{i} = sdpvar(d,d,type,'complex');
    x = x + kron(xpart{i},u{i});
    if can_herm{i} == 0 % Non-hermitian unknown
        x = x + kron(xpart{i}',u{i}');
    elseif can_herm{i} == 2 % Observable
        val = val + trace((ineqops(:,:,can_outcome{i},can_setting{i}) ...
                         - ineqops(:,:,outcomes,can_setting{i})) * u{i});
    elseif can_herm{i} == 3 % Identity, i.e. reduced state
        val = val + trace(sum(ineqops(:,:,outcomes,:),4)*u{i});
    end
end

F = [trace(u{1}) == 1, x >= 0, Tx(x,2,[outsize,d]) >= 0];
opts = sdpsettings('verbose', 0);

optimize(F, -val,opts)
maxval = value(val);

optimize(F, val,opts)
minval = value(val);
