% CHI_VALIDITY   Returns the matrices needed to enforce chi-validity
%
%  Based on <a href="http://arxiv.org/abs/1305.1767">arXiv:1305.1767</a>
%
%  [outsize,xpart,num_can,can_outcome,can_setting,can_herm] = ...
%       CHI_VALIDITY(outcomes, settings, level) takes the parameters:
%
%   outcomes, aka $n_A$: Number of outcomes from untrusted system (Alice)
%   settings, aka $m_A$: Number of settings for untrusted system (Alice)
%   level, aka $l$: Level of hierarchy
%
%  And returns the following:
% 
%   outsize: Block dimension of chi-matrix = (settings*(outcomes-1) + 1)^level
%   xpart: Matrices of 1s and 0s that build up the block-matrix form of chi
%   num_can: Number of such matrices
%   can_outcome: The measurement outcome associated with an xpart, if any
%   can_setting: The settings associated with an xpart, if any
%   can_herm: The type of block in chi where the relevant xpart is 1:
%     0: Unknown arbitrary d by d matrix
%     1: Unknown Hermitian d by d matrix
%     2: Bob's collapsed state sigma_{can_outcome | can_setting}
%     3: Bob's reduced state
%
%  Latest version at
%  <a href="https://www.mattpusey.uk/steer/">www.mattpusey.co.uk/steer</a>,
%  bug reports welcome at <a href="mailto:m@physics.org">m@physics.org</a>

% Copyright (C) 2013 Matthew F. Pusey
%
% This program is free software; you can redistribute it and/or
% modify it under the terms of the GNU General Public License
% as published by the Free Software Foundation; either version 2
% of the License, or (at your option) any later version.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
%
% You should have received a copy of the GNU General Public License
% along with this program; if not, write to the Free Software
% Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
% MA 02110-1301, USA.

function [Msize,xpart,num_can,can_outcome,can_setting,can_herm] ...
    = chi_validity(outcomes, settings, level)


fn = fullfile('chivalidititycache', strcat(...
    int2str(outcomes),'_',int2str(settings), '_', int2str(level)));
try
load(fn);
catch ME % Presumably new parameters, so calculate and save into cache
    disp('creating new chivaliditycache entry...');
    [Msize,xpart,num_can,can_outcome,can_setting,can_herm] ...
        = chi_validitity_full(outcomes, settings, level);
    if ~exist('chivalidititycache', 'dir')
        mkdir('chivalidititycache');
    end
    save(fn,'Msize','xpart','num_can','can_outcome','can_setting','can_herm');
    disp('...done');
end

function [outsize,xpart,num_can,can_outcome,can_setting,can_herm] ...
    = chi_validitity_full(outcomes, settings, level)
% Actually do calculation, used for cache miss

Msize = settings*(outcomes-1) + 1;
outsize = Msize^level;
powers = Msize.^(0:(2*level)-1);

num_can = 0;
for possible=0:(outsize)^2-1
    string = floor(mod(possible./powers,Msize));
    string = string(1:find(string, 1, 'last'));
    len = length(string);
    if len == 0
        string = [0];
        len = 1;
    end
    canstring = applyidentities(string, outcomes);
    if ((len ~= length(canstring) || any(string ~= canstring)))
        continue
    end
    halflength = floor(len/2);
    if (halflength > 0) % Unobservable quantity
        h1 = string(1:halflength);
        h2 = string(len-halflength+1:len);
        h1 = sum(powers(1:halflength) .* h1);
        h2 = sum(powers(halflength:-1:1) .* h2);
        if (h1 == h2)
            hermitian = 1;
        elseif (h1 > h2)
            continue
        else
            hermitian = 0;
        end
    elseif string % Observable
        hermitian = 2;
    else % Identity
        hermitian = 3;
    end
    num_can = num_can + 1;
    can{num_can} = canstring;
    can_herm{num_can} = hermitian;
    xpart{num_can} = zeros(outsize);
    if (hermitian == 2)
        setting = floor((canstring(1)-1)/(outcomes-1)) + 1;
        outcome = canstring(1) - (setting-1)*(outcomes-1);
        can_setting{num_can} = setting;
        can_outcome{num_can} = outcome;
    end
end
powers = Msize.^(level-1:-1:0);
revpowers = Msize.^(0:level-1);
for row = 0:outsize-1
    rowstring = floor(mod(row./powers,Msize));
    for col = 0:outsize-1
        string = [floor(mod(col./revpowers,Msize)), rowstring];
        canstring = applyidentities(string, outcomes);
        for i = 1:num_can
            if length(can{i}) == length(canstring) && all(can{i} == canstring)
                xpart{i}(row+1,col+1) = 1;
            end
        end
    end
end

function canonicalstring = applyidentities(string, outcomes)
% Takes a string of representing a product of operators and simplifies by
% applying the following identities:
%
% [x, 0, y] = [x, y] because 0 represents an identity operator
% [x, x] = [x] because projectors square to themselves
% [x, y] = [] if x and y are different outcomes of same measurement
%
% Conventions: 0 means final string is corresponds to identity operator
% [] means final string corresponds to zero operator

oldindex = -1;
canonicalstring = [];
if string == 0 %% All identities
    canonicalstring = 0;
    return
end
for index = string
    if index == 0 || index == oldindex % Identity || repeated projector
        continue
    elseif floor((index-1)/(outcomes-1)) == floor((oldindex-1)/(outcomes-1)) % Orthogonality when setting equal
        canonicalstring = [];
        return
    else
        canonicalstring = [canonicalstring, index]; %#ok<AGROW>
    end
    oldindex = index;
end
