% MINNEG Calculate lower bound on the negativity to produce assemblage
%
%  Based on <a href="http://arxiv.org/abs/1305.1767">arXiv:1305.1767</a>, eq. 7
%
%  val = MINNEG(assemblage, level) takes the parameters:
%
%   assemblage: size [$d_B$, $d_B$, $n_A$, $m_A$], collection of sigmas
%   level, aka $l$: Level of hierarchy
%
%  And attempts to solve Equation (7) in paper.
%
%  Requires chi_validity.m, Tx.m, and a working YALMIP setup
%
%  Latest version at
%  <a href="https://www.mattpusey.uk/steer/">www.mattpusey.co.uk/steer</a>,
%  bug reports welcome at <a href="mailto:m@physics.org">m@physics.org</a>
%
%  Thanks to Shin-Liang Chen for identifying a bug in an earlier version.

% Copyright (C) 2013-2018 Matthew F. Pusey
%
% This program is free software; you can redistribute it and/or
% modify it under the terms of the GNU General Public License
% as published by the Free Software Foundation; either version 2
% of the License, or (at your option) any later version.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
%
% You should have received a copy of the GNU General Public License
% along with this program; if not, write to the Free Software
% Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
% MA 02110-1301, USA.

function val = minneg(assemblage,level)
[~,d,outcomes,settings] = size(assemblage);
[outsize,xpart,num_can,can_outcome,can_setting,can_herm] ...
    = chi_validity(outcomes, settings, level);

xsize = d*outsize;
xp = zeros(xsize); % Denoted $\chi_+^{T_B}$ in paper
xm = zeros(xsize); % Denoted $\chi_-^{T_B}$ in paper
F = [];
up = cell(num_can);
um = cell(num_can);

for i=1:num_can
    if can_herm{i}
        type = 'hermitian';
    else
        type = 'full';
    end
    up{i} = sdpvar(d,d,type,'complex');
    um{i} = sdpvar(d,d,type,'complex');
    xp = xp + kron(xpart{i},up{i});
    xm = xm + kron(xpart{i},um{i});
    if can_herm{i} == 0 % Non-hermitian unknown
        xp = xp + kron(xpart{i}',up{i}');
        xm = xm + kron(xpart{i}',um{i}');
    elseif can_herm{i} == 2 % Observable
        F = F + [up{i} - um{i} == ...
            assemblage(:,:,can_outcome{i},can_setting{i}).'];
    elseif can_herm{i} == 3 % Identity, i.e. reduced state
        neg = trace(um{i});
        F = F + [up{i} - um{i} == ...
            sum(assemblage(:,:,:,1),3).'];
    end
end

F = F + [xp >= 0, xm >= 0, Tx(xp - xm,2,[outsize,d]) >= 0];
optimize(F,neg)
val = value(neg);
