% MINNEG_INEQ Calculate lower bound on the negativity to violate ineq
%
%  Based on <a href="http://arxiv.org/abs/1305.1767">arXiv:1305.1767</a>, eq. 8
%
%  val = MINNEG_INEQ(ineqops,ineqvalues,level) takes the parameters:
%
%   ineqops: size [$d_B$, $d_B$, $n_A$, $m_A$], collection of F matrices
%   ineqvalues: array of values for F
%   level, aka $l$: Level of hierarchy
%
%  And repeatedly attempts to solve Equation (8) in paper, once for each
%  ineqvalue, returning an array of the corresponding negativity bounds
%
%  Requires chi_validity.m, Tx.m, and a working YALMIP setup
%
%  Latest version at
%  <a href="https://www.mattpusey.uk/steer/">www.mattpusey.co.uk/steer</a>,
%  bug reports welcome at <a href="mailto:m@physics.org">m@physics.org</a>

% Copyright (C) 2013-2018 Matthew F. Pusey
%
% This program is free software; you can redistribute it and/or
% modify it under the terms of the GNU General Public License
% as published by the Free Software Foundation; either version 2
% of the License, or (at your option) any later version.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
%
% You should have received a copy of the GNU General Public License
% along with this program; if not, write to the Free Software
% Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
% MA 02110-1301, USA.

function val = minneg_ineq(ineqops,ineqvalues,level)
[~,d,outcomes,settings] = size(ineqops);
n = length(ineqvalues);
[outsize,xpart,num_can,can_outcome,can_setting,can_herm] ...
    = chi_validity(outcomes, settings, level);

xsize = d*outsize;
xp = zeros(xsize); % Denoted $\chi_+^{T_B}$ in paper
xm = zeros(xsize); % Denoted $\chi_+^{T_B}$ in paper
thisval = 0;
up = cell(num_can);
um = cell(num_can);

for i=1:num_can
    if can_herm{i}
        type = 'hermitian';
    else
        type = 'full';
    end
    up{i} = sdpvar(d,d,type,'complex');
    um{i} = sdpvar(d,d,type,'complex');
    xp = xp + kron(xpart{i},up{i});
    xm = xm + kron(xpart{i},um{i});
    if can_herm{i} == 0 % Non-hermitian unknown
        xp = xp + kron(xpart{i}',up{i}');
        xm = xm + kron(xpart{i}',um{i}');
    elseif can_herm{i} == 2 % Observable
        thisval = thisval + trace((ineqops(:,:,can_outcome{i},can_setting{i}) ...
                         - ineqops(:,:,outcomes,can_setting{i})).' * (up{i}-um{i}));
    elseif can_herm{i} == 3 % Identity, i.e. reduced state
        neg = trace(um{i});
        norm = trace(up{i} - um{i});
        thisval = thisval + trace(sum(ineqops(:,:,outcomes,:),4).' * (up{i}-um{i}));
    end
end

val = zeros(1,n);
opts = sdpsettings('verbose', 0);
for i=1:n
    F = [norm == 1, thisval == ineqvalues(i), xp >= 0, xm >= 0, Tx(xp-xm,2,[outsize,d]) >= 0];
    optimize(F,neg,opts)
    val(i) = value(neg);
end
